/*
  @copyright Steve Keen 2012
  @author Russell Standish
  This file is part of Minsky.

  Minsky is free software: you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Minsky is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Minsky.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "plotWidget.h"
#include "variable.h"
#include "init.h"
#include "cairoItems.h"
#include "minsky.h"
using namespace ecolab::cairo;
using namespace ecolab;
using namespace std;

namespace
{
  const int numLines = 4; // number of simultaneous variables to plot

  map<string, shared_ptr<TkPhotoSurface> > surfaces;

  Plots::Map& plots() {return minsky.plots.plots;}

  struct PlotItem: public cairo::CairoImage
  {
    PlotWidget& pw;
    string image;
    PlotItem(const std::string& image): 
      CairoImage(image), pw(plots()[image]), image(image)  {
      surfaces[image]=cairoSurface;
      if (cairoSurface && pw.images.empty())
        {      
          pw.images.push_back(image);
          float w=cairoSurface->width(), h=cairoSurface->height();
          float x = -0.5*w; // x location of ports
          float dy = h/(numLines); // distance between ports
          pw.x=0; pw.y=0; // set up everything relative to pw centroid
          for (float y=0.5*(dy-h); y<0.5*h; y+=dy)
            pw.ports<<=portManager().addPort(Port(x,y,true));

          // add in the x variable ports
          float y=-0*h, dx=w/numLines;
          for (float x=0.5*(dx-w); x<0.5*w; x+=dx)
             pw.ports<<=portManager().addPort(Port(x,y,true));

        }
      pw.nTicks=3;
      pw.fontScale=5;
      pw.offx=10; pw.offy=5;
      pw.Image(image);
    }
    ~PlotItem() {
      plots().erase(image);
    }
    void draw();
  };

  void PlotItem::draw()
  {
    if (cairoSurface)
      {
        cairoSurface->clear();
        cairo_t* cairo=cairoSurface->cairo();
        double w=cairoSurface->width(), h=cairoSurface->height();

        //        cairo_save(cairo);
        cairo_identity_matrix(cairo);

        cairo_reset_clip(cairo);
        cairo_new_path(cairo);
        cairo_rectangle(cairo,0,0,w,h);
        cairo_clip(cairo);

        // draw ports
        for (size_t i=0; i<pw.ports.size(); ++i)
          drawTriangle(cairo, portManager().ports[pw.ports[i]].x-pw.x+0.5*w, 
                       portManager().ports[pw.ports[i]].y-pw.y+0.5*h, 
                       palette[(i%numLines)%paletteSz],
                       (i<numLines)? 0: -0.5*M_PI);


        pw.draw(*cairoSurface);

        cairoSurface->blit();
        //        cairo_restore(cairo);
      }
  }

#ifdef __GNUC__
#pragma GCC push
#pragma GCC diagnostic ignored "-Wwrite-strings"
#endif
  // register PlotItem with Tk for use in canvases.
  int registerPlotItem()
  {
    static Tk_ItemType plotItemType = cairoItemType;
    plotItemType.name="plot";
    plotItemType.createProc=createImage<PlotItem>;
    Tk_CreateItemType(&plotItemType);
    return 0;
  }
#ifdef __GNUC__
#pragma GCC pop
#endif

  /// temporarily sets nTicks and fontScale, restoring them on scope exit
  struct SetTicksAndFontSize
  {
    int nTicks;
    double fontScale;
    PlotWidget& p;
    SetTicksAndFontSize(PlotWidget& p, int n, double f):
      p(p), nTicks(p.nTicks), fontScale(p.fontScale) 
    {
      p.nTicks=n;
      p.fontScale=f;
    }
    ~SetTicksAndFontSize()
    {
      p.nTicks=nTicks;
      p.fontScale=fontScale;
    }
  };
}

//static int reg=registerPlotItem();
static int reg=(initVec().push_back(registerPlotItem), 0);

/// TODO: better handled in a destructor, but for the moment, the
/// software architecture requires that this be called explicitly from
/// Minsky::deletePlot
void PlotWidget::deletePorts()
{
  for (size_t p=0; p<ports.size(); ++p)
    portManager().delPort(ports[p]);
  ports.resize(0);
}

void PlotWidget::moveTo(float x1, float y1)
{
  float w=width(), h=height();
  float dx=x1-x, dy=y1-y;
  x=x1; y=y1;
  {
    float x = x1-0.5*w; // x location of ports
    float dy = h/(numLines); // distance between ports
    float y=y1+0.5*(dy-h);
    size_t i=0;
    for (; i<numLines; ++i, y+=dy)
      portManager().movePortTo(ports[i],x,y);
    float dx=w/numLines;
    x=x1+0.5*(dx-w);
    y=y1+0.5*h;
    for (; i<2*numLines; ++i, x+=dx)
      portManager().movePortTo(ports[i],x,y);
  }
}

void PlotWidget::addPlotPt(double t)
{
  for (size_t i=0; i<images.size(); ++i)
    {
      map<string, shared_ptr<TkPhotoSurface> >::iterator surf=
        surfaces.find(images[i]);
      if (surf!=surfaces.end())
        {
          SetTicksAndFontSize stf(*this, i==0? nTicks:10, i==0? fontScale:1);
          for (size_t pen=0; pen<yvars.size(); ++pen)
            if (yvars[pen].idx()>=0)
              switch (xvars.size())
                {
                case 0: // use t, when x variable not attached
                  add(*surf->second, pen, t, yvars[pen].value());
                  break;
                case 1: // use the value of attached variable
                  assert(xvars[0].idx()>=0);
                  add(*surf->second, pen, xvars[0].value(), yvars[pen].value());
                  break;
                default:
                  if (pen < xvars.size() && xvars[pen].idx()>=0)
                    add(*surf->second, pen, xvars[pen].value(), yvars[pen].value());
                  else
                    throw error("x input not wired for pen %d",(int)pen+1);
                  break;
                }
                  
          //TODO: I honestly do not know why these lines need to be here.
          surf->second->clear();
          draw(*surf->second);
          surf->second->blit();
          
        }
    }
}

static VariableValue disconnected;

void PlotWidget::connectVar(const VariableValue& var, int pen)
{
  if (pen>=0)
    if (pen<numLines)
      {
        yvars.resize(pen+1);
        yvars[pen]=var;
      }
    else if (pen<2*numLines)
      {
        xvars.resize(pen-numLines+1);
        xvars[pen-numLines]=var;
      }
}

void PlotWidget::disconnectPen(int pen)
{
  if (pen<yvars.size())
    yvars[pen]=VariableValue();
  if (pen<xvars.size())
    xvars[pen]=VariableValue();
}

void Plots::moveTo(TCL_args args)
{
  plots[args[0]].moveTo(args[1],args[2]);
}

void Plots::addImage(TCL_args args) {
  string id=args, image=args;
  surfaces[image].reset
    (new TkPhotoSurface(Tk_FindPhoto(interp(), image.c_str()), false));
  PlotWidget& p=plots[id];
  p.images.push_back(image);
  // save & restore nTicks and fontscale
  int nTicks=p.nTicks;
  double fontScale=p.fontScale;
  SetTicksAndFontSize stf(p,10,1);
  p.draw(*surfaces[image]);
  surfaces[image]->blit();
}

void Plots::reset(const VariableManager& vm)
{
  for (Map::iterator p=plots.begin(); p!=plots.end(); ++p)
    p->second.clear();
}

string Plots::nextPlotID()
{
  int id=0;
  if (!plots.empty())
    {
      string lastID=plots.rbegin()->first;
      sscanf(lastID.c_str(),"plot_image%d",&id);
      id++;
    }
  ostringstream os;
  os << "plot_image"<<id;
  return os.str();
}
